from .losses import calculate_bg_spp_loss, calculate_bg_spp_loss_normal, get_logp_boundary
from .losses import normal_fl_weighting, abnormal_fl_weighting


__all__ = ['calculate_bg_spp_loss',
           'calculate_bg_spp_loss_normal',
           'get_logp_boundary',
           'normal_fl_weighting',
           'abnormal_fl_weighting']


import os
import numpy as np
import torch
from .config import parse_args
from .utils.utils import init_seeds, setting_lr_parameters


def main_single(device,gpu,seed,input_size,batch_size,lr,meta_epochs,sub_epochs,train_list,val_list,save_model_path):
    
    init_seeds(seed)
    args = parse_args()
    if device == 'GPU':
        os.environ['CUDA_VISIBLE_DEVICES'] = gpu
        args.device = torch.device("cuda")
    else:
        args.device = torch.device("cpu")
    # model path
    # args.model_path = "{}_{}_{}_{}".format(
    #     args.dataset, args.backbone_arch, args.flow_arch, args.class_name)
    args.model_path = save_model_path
    # image
    # args.img_size = (input_size, input_size)  
    # args.crop_size = (input_size, input_size)
    args.img_size = input_size  
    args.crop_size = input_size
      
    args.norm_mean, args.norm_std = [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
    
    args.img_dims = [3] + list(args.img_size)
    # params
    args.batch_size = batch_size
    args.lr = lr
    args.meta_epochs = meta_epochs
    args.sub_epochs = sub_epochs


    # output settings
    args.save_results = True
    # unsup-train lr settings
    setting_lr_parameters(args)
    
    ############################################################################################################################################################
    # set data
    args.train_list = train_list
    args.val_list = val_list
    ############################################################################################################################################################

    # selecting train functions
    if args.with_fas:
        from .engines.bgad_fas_train_engine import train
        img_auc, pix_auc = train(args)
    else:
        from .engines.bgad_train_engine import train
        img_auc, pix_auc = train(args)

    return img_auc, pix_auc





from hivesda import test
import os

# # setting cuda

device = "GPU"   # CPU or GPU
gpu = "0"        # set gpu numer
input_size = 300 # set image size e.g.) 300 -> (300x300)    
run_model_path = "/workspace/output_custom_list/bgad_fas_10/weights/custom_tf_efficientnet_b6_conditional_flow_model_bottle.pt"
save_heatmap_path = "./vis_result"
######################################################################################################################################################################
# set data
train_list = []
val_list = []
ng_list = []

val_normal_list = os.listdir("/workspace/Data_toptec_center_png300/test/good")
for v in val_normal_list:
    val_list.append(["/workspace/Data_toptec_center_png300/test/good/"+v,0,None,'good'])

total_ng_list = os.listdir("/workspace/Data_toptec_center_png300/test/NG")
for n in total_ng_list:
    ng_list.append(["/workspace/Data_toptec_center_png300/test/NG/"+n,1,"/workspace/Data_toptec_center_png300/ground_truth/NG/"+n.split(".")[0]+"_mask.png",'NG'])

val_list = val_list + ng_list
######################################################################################################################################################################



img_auc, pix_auc, img_scores = test.main_single(device,gpu,input_size,val_list,run_model_path,save_heatmap_path)
print(f'Image-AUC: {img_auc}, Pixel-AUC: {pix_auc}')
print(img_scores)





import torch
import math
import numpy as np

def lerp_np(x,y,w):
    fin_out = (y-x)*w + x
    return fin_out

def generate_fractal_noise_2d(shape, res, octaves=1, persistence=0.5):
    noise = np.zeros(shape)
    frequency = 1
    amplitude = 1
    for _ in range(octaves):
        noise += amplitude * generate_perlin_noise_2d(shape, (frequency*res[0], frequency*res[1]))
        frequency *= 2
        amplitude *= persistence
    return noise


def generate_perlin_noise_2d(shape, res):
    def f(t):
        return 6 * t ** 5 - 15 * t ** 4 + 10 * t ** 3

    delta = (res[0] / shape[0], res[1] / shape[1])
    d = (shape[0] // res[0], shape[1] // res[1])
    grid = np.mgrid[0:res[0]:delta[0], 0:res[1]:delta[1]].transpose(1, 2, 0) % 1
    # Gradients
    angles = 2 * np.pi * np.random.rand(res[0] + 1, res[1] + 1)
    gradients = np.dstack((np.cos(angles), np.sin(angles)))
    g00 = gradients[0:-1, 0:-1].repeat(d[0], 0).repeat(d[1], 1)
    g10 = gradients[1:, 0:-1].repeat(d[0], 0).repeat(d[1], 1)
    g01 = gradients[0:-1, 1:].repeat(d[0], 0).repeat(d[1], 1)
    g11 = gradients[1:, 1:].repeat(d[0], 0).repeat(d[1], 1)
    # Ramps
    n00 = np.sum(grid * g00, 2)
    n10 = np.sum(np.dstack((grid[:, :, 0] - 1, grid[:, :, 1])) * g10, 2)
    n01 = np.sum(np.dstack((grid[:, :, 0], grid[:, :, 1] - 1)) * g01, 2)
    n11 = np.sum(np.dstack((grid[:, :, 0] - 1, grid[:, :, 1] - 1)) * g11, 2)
    # Interpolation
    t = f(grid)
    n0 = n00 * (1 - t[:, :, 0]) + t[:, :, 0] * n10
    n1 = n01 * (1 - t[:, :, 0]) + t[:, :, 0] * n11
    return np.sqrt(2) * ((1 - t[:, :, 1]) * n0 + t[:, :, 1] * n1)


def rand_perlin_2d_np(shape, res, fade=lambda t: 6 * t ** 5 - 15 * t ** 4 + 10 * t ** 3):
    delta = (res[0] / shape[0], res[1] / shape[1])
    d = (shape[0] // res[0], shape[1] // res[1])
    grid = np.mgrid[0:res[0]:delta[0], 0:res[1]:delta[1]].transpose(1, 2, 0) % 1

    angles = 2 * math.pi * np.random.rand(res[0] + 1, res[1] + 1)
    gradients = np.stack((np.cos(angles), np.sin(angles)), axis=-1)
    tt = np.repeat(np.repeat(gradients,d[0],axis=0),d[1],axis=1)

    tile_grads = lambda slice1, slice2: np.repeat(np.repeat(gradients[slice1[0]:slice1[1], slice2[0]:slice2[1]],d[0],axis=0),d[1],axis=1)
    dot = lambda grad, shift: (
                np.stack((grid[:shape[0], :shape[1], 0] + shift[0], grid[:shape[0], :shape[1], 1] + shift[1]),
                            axis=-1) * grad[:shape[0], :shape[1]]).sum(axis=-1)

    n00 = dot(tile_grads([0, -1], [0, -1]), [0, 0])
    n10 = dot(tile_grads([1, None], [0, -1]), [-1, 0])
    n01 = dot(tile_grads([0, -1], [1, None]), [0, -1])
    n11 = dot(tile_grads([1, None], [1, None]), [-1, -1])
    t = fade(grid[:shape[0], :shape[1]])
    return math.sqrt(2) * lerp_np(lerp_np(n00, n10, t[..., 0]), lerp_np(n01, n11, t[..., 0]), t[..., 1])


def rand_perlin_2d(shape, res, fade=lambda t: 6 * t ** 5 - 15 * t ** 4 + 10 * t ** 3):
    delta = (res[0] / shape[0], res[1] / shape[1])
    d = (shape[0] // res[0], shape[1] // res[1])

    grid = torch.stack(torch.meshgrid(torch.arange(0, res[0], delta[0]), torch.arange(0, res[1], delta[1])), dim=-1) % 1
    angles = 2 * math.pi * torch.rand(res[0] + 1, res[1] + 1)
    gradients = torch.stack((torch.cos(angles), torch.sin(angles)), dim=-1)

    tile_grads = lambda slice1, slice2: gradients[slice1[0]:slice1[1], slice2[0]:slice2[1]].repeat_interleave(d[0],
                                                                                                              0).repeat_interleave(
        d[1], 1)
    dot = lambda grad, shift: (
                torch.stack((grid[:shape[0], :shape[1], 0] + shift[0], grid[:shape[0], :shape[1], 1] + shift[1]),
                            dim=-1) * grad[:shape[0], :shape[1]]).sum(dim=-1)

    n00 = dot(tile_grads([0, -1], [0, -1]), [0, 0])

    n10 = dot(tile_grads([1, None], [0, -1]), [-1, 0])
    n01 = dot(tile_grads([0, -1], [1, None]), [0, -1])
    n11 = dot(tile_grads([1, None], [1, None]), [-1, -1])
    t = fade(grid[:shape[0], :shape[1]])
    return math.sqrt(2) * torch.lerp(torch.lerp(n00, n10, t[..., 0]), torch.lerp(n01, n11, t[..., 0]), t[..., 1])


def rand_perlin_2d_octaves(shape, res, octaves=1, persistence=0.5):
    noise = torch.zeros(shape)
    frequency = 1
    amplitude = 1
    for _ in range(octaves):
        noise += amplitude * rand_perlin_2d(shape, (frequency * res[0], frequency * res[1]))
        frequency *= 2
        amplitude *= persistence
    return noise


import torch
from .custom_list import CustomDataset,CustomFSCopyPasteDataset
from .utils import BalancedBatchSampler

__all__ = ['CustomDataset',
           'CustomFSCopyPasteDataset',
           'create_data_loader',
           'create_fas_data_loader',
           'create_test_data_loader']


def create_data_loader(args):
    kwargs = {'num_workers': args.num_workers, 'pin_memory': True}
    if args.dataset == 'custom':
        train_dataset = CustomDataset(args, is_train=True)
        test_dataset  = CustomDataset(args, is_train=False)
    else:
        raise NotImplementedError('{} is not supported dataset!'.format(args.dataset))
    # dataloader
    train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True, drop_last=True, **kwargs)
    test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=1, shuffle=False, drop_last=False, **kwargs)

    return train_loader, test_loader


def create_fas_data_loader(args):
    kwargs = {'num_workers': args.num_workers, 'pin_memory': True}
    if args.dataset == 'custom':
        normal_dataset = CustomDataset(args, is_train=True)
        if args.data_strategy == '0':
            # train_dataset = MVTecFSDataset(args, is_train=True)
            pass
        elif args.data_strategy == '0,1':
            train_dataset = CustomFSCopyPasteDataset(args, is_train=True)
        elif args.data_strategy == '0,2':
            # train_dataset = MVTecPseudoDataset(args, is_train=True)
            input("check data_strategy : 0,2")
            pass
        elif args.data_strategy == '0,1,2':
            # train_dataset = MVTecAnomalyDataset(args, is_train=True)
            input("check data_strategy : 0,1,2")
            pass
        if args.not_in_test:
            test_dataset  = CustomDataset(args, is_train=False, excluded_images=train_dataset.a_imgs)
        else:
            test_dataset  = CustomDataset(args, is_train=False)
    else:
        raise NotImplementedError('{} is not supported dataset!'.format(args.dataset))
    # dataloader
    normal_loader = torch.utils.data.DataLoader(normal_dataset, batch_size=args.batch_size, shuffle=True, drop_last=True, **kwargs)
    if args.balanced_data_loader:
        train_loader = torch.utils.data.DataLoader(train_dataset, batch_sampler=BalancedBatchSampler(args, train_dataset), **kwargs)
    else:
        train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True, drop_last=True, **kwargs)
    test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=1, shuffle=False, drop_last=False, **kwargs)

    return normal_loader, train_loader, test_loader


def create_test_data_loader(args):
    kwargs = {'num_workers': args.num_workers, 'pin_memory': True}
    if args.dataset == 'custom':
        test_dataset  = CustomDataset(args, is_train=False)
    else:
        raise NotImplementedError('{} is not supported dataset!'.format(args.dataset))
    # dataloader
    test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=1, shuffle=False, drop_last=False, **kwargs)

    return test_loader